using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

namespace MVVMDemo.ViewModel
{

  /// <summary>
  /// ViewModel fr einen Datums-Wert (ohne Zeit).
  /// Merkt sich den Original-Wert fr das UndoChanges().
  /// Zustzlich zum Value gibt einen DisplayValue und einen EditValue.
  /// </summary>
  public class DateViewModel : ViewModelBase, IComparable
  {
    private DateTime? _value;
    private DateTime? _originalValue;
    private bool _inEdit;

    public DateViewModel(DateTime? value)
    {
      if (value.HasValue && value.Value != value.Value.Date)
      {
        throw new ArgumentOutOfRangeException("value", "DateTime value must be a pure date.");
      }
      _value = value;
      _originalValue = value;
      if (!_value.HasValue) ErrorMessage = "Invalid date value."; 
    }

    // Der eigentliche Wert.
    public DateTime Value
    {
      get { return _value.Value; }
      set {
        DateTime dateOnly = value.Date;
        if (dateOnly != _value)
        {
          _value = dateOnly;
          RaisePropertyChangedAll();
          HasChanges = _value != _originalValue;
        }
      }
    }

    // Langes Datumsformat fr die Anzeige
    public string DisplayValue
    {
      get
      {
        if (this.HasError) return "???";
        if (!_value.HasValue) return string.Empty;
        return _value.Value.ToLongDateString();
      }
    }

    // Kurzes Datumsformat fr das manuelle Editieren
    public string EditValue
    {
      get
      {
        if (!_value.HasValue || this.HasError) return string.Empty;
        return _value.Value.ToShortDateString();
      }
      set
      {
        _inEdit = true;
        try
        {
          DateTime newval;
          if (DateTime.TryParse(value, out newval))
          {
            HasError = false;
            Value = newval;
          }
          else
          {
            ErrorMessage = "Invalid date format.";
          }
          RaisePropertyChangedAll();
        }
        finally
        {
          _inEdit = false;
        }
      }
    }

    public override void AcceptChanges()
    {
      _originalValue = _value;
      HasChanges = false;
      HasError = false;
      RaisePropertyChangedAll();
    }

    public override void UndoChanges()
    {
      _value = _originalValue;
      HasError = false;
      RaisePropertyChangedAll();
    }

    // Muss implementiert werden, damit die Sortierung im CollectionView funktioniert.
    public int CompareTo(object obj)
    {
      if (obj == null) return 1;
      DateViewModel other = obj as DateViewModel;
      if (other == null) throw new ArgumentOutOfRangeException("obj", "Must be of type DateViewModel");
      return ((IComparable)_value).CompareTo(other._value);
    }

    // Anzeige fr alle Values aufrischen 
    private void RaisePropertyChangedAll()
    {
      RaisePropertyChanged("Value");
      RaisePropertyChanged("DisplayValue");
      if (!_inEdit) RaisePropertyChanged("EditValue");
    }

  }
}
