using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Windows.Data;
using System.Windows.Input;
using MVVMDemo.Model;
using System.Windows;

namespace MVVMDemo.ViewModel
{
  /// <summary>
  /// Diese ViewModel Klasse korrespondiert mit dem Haupt-Fenster der Applikation und 
  /// enthlt den grssten Teil der Applikations-Logik. PersonAppViewModel enthlt unter anderem die Daten,
  /// also die Liste der Personen und die CommandBindings fr New, Delete, Save. 
  /// </summary>
  public class PersonAppViewModel : ViewModelBase
  {

    #region Private Felder

    // Fr die ComboBox zur Auswahl der Sortierung
    private static string[] s_sortSelection = {"Firstname", "Lastname", "Birthdate"};

    // Referenz auf die IPersonDataProvider Instanz
    private IPersonDataProvider _dataProvider;

    // Collection mit ViewModels fr Personen
    private ObservableCollection<PersonViewModel> _persons;

    // CollectionView auf Personen
    private ListCollectionView _personsView;

    // Aktuell ausgewhlte Sortierung
    private string _sortByProperty = s_sortSelection[0];

    // Personen, die gelscht wurden (werden erst beim Save wirklich gelscht)
    private List<PersonViewModel> _deletedPersons = new List<PersonViewModel>();

    // CommandBindings
    private CommandBinding _newPersonCommandBinding;
    private CommandBinding _deletePersonCommandBinding;
    private CommandBinding _savePersonsCommandBinding;

    #endregion

    #region Konstruktor

    public PersonAppViewModel(IPersonDataProvider dataProvider)
    {
      _dataProvider = dataProvider;

      _persons = new ObservableCollection<PersonViewModel>();
      _personsView = new ListCollectionView(_persons);
      
      PersonsView.CurrentChanged += delegate
      {
        // IsEditEnabled kann sich ndern wenn die angewhlte Person sich ndert
        RaisePropertyChanged("IsEditEnabled");
      };

      this.PropertyChanged += delegate(object sender, PropertyChangedEventArgs e)
      {
        // Fenster-Titel ndert sich, wenn HasChanges sich ndert (das Sternchen)  
        if(e.PropertyName == "HasChanges") RaisePropertyChanged("WindowTitle");
      };

      // Erzeugen der CommandBindings und Verknpfung mit den Handler-Methoden
      _newPersonCommandBinding = new CommandBinding(ApplicationCommands.New, HandleNewPersonExecuted, HandleNewPersonCanExecute);
      _deletePersonCommandBinding = new CommandBinding(ApplicationCommands.Delete, HandleDeletePersonExecuted, HandleDeletePersonCanExecute);
      _savePersonsCommandBinding = new CommandBinding(ApplicationCommands.Save, HandleSavePersonsExecuted, HandleSavePersonsCanExecute);
    }

    #endregion

    #region Diverse Properties (WindowTitle, PersonsView, SortSelection, SortByProperty, IsEditEnabled)

    public string WindowTitle
    {
      get
      {
        string title = "M-V-VM Demo Application";
        if (this.HasChanges)
        {
          title += " *";
        }
        return title;
      }

    }

    public ListCollectionView PersonsView
    {
      get { return _personsView; }
    }

    public IEnumerable<string> SortSelection
    {
      get { return s_sortSelection; }
    }

    public string SortByProperty
    {
      get { return _sortByProperty; }
      set {
        if (value != _sortByProperty)
        {
          _sortByProperty = value;
          UpdateSorting();
          RaisePropertyChanged("SortByProperty");
        }
      }
    }

    public bool IsEditEnabled
    {
      get
      {
        if (_personsView == null) return false;
        return _personsView.CurrentItem != null;
      }
    }

    #endregion

    #region Command Handling

    //
    // New Command:
    //
    public CommandBinding NewPersonCommandBinding
    {
      get { return _newPersonCommandBinding; }
    }

    public void HandleNewPersonCanExecute(object sender, CanExecuteRoutedEventArgs e)
    {
      e.CanExecute = true;
    }    
    
    public void HandleNewPersonExecuted(object sender, ExecutedRoutedEventArgs e)
    {
      InternalInsertPerson();
    }

    //
    // Delete Command:
    //

    public CommandBinding DeletePersonCommandBinding
    {
      get { return _deletePersonCommandBinding; }
    }

    public void HandleDeletePersonCanExecute(object sender, CanExecuteRoutedEventArgs e)
    {
      if(PersonsView.CurrentItem != null) e.CanExecute = true;
    }

    public void HandleDeletePersonExecuted(object sender, ExecutedRoutedEventArgs e)
    {
      PersonViewModel personVM = PersonsView.CurrentItem as PersonViewModel;
      if (personVM != null)
      {
        string msg = string.Format("Really delete person '{0} {1}'?", personVM.Firstname, personVM.Lastname);
        UserFeedbackEventArgs ufe =
          new UserFeedbackEventArgs("Delete", msg, MessageBoxButton.YesNo, MessageBoxImage.Question);
        RaiseUserFeedbackRequired(ufe);
        if (ufe.Result == MessageBoxResult.Yes)
        {
          InternalDeletePerson(personVM);
        }
      }
    }
 
    //
    // Save Command:
    //
    public CommandBinding SavePersonsCommandBinding
    {
      get { return _savePersonsCommandBinding; }
    }

    public void HandleSavePersonsCanExecute(object sender, CanExecuteRoutedEventArgs e)
    {
      e.CanExecute = this.HasChanges && !this.HasError;
    }

    public void HandleSavePersonsExecuted(object sender, ExecutedRoutedEventArgs e)
    {
      this.SendChangesToModel();
      this.AcceptChanges();
    }

    #endregion

    #region ffentliche Methoden

    // Laden der Personen ber den Data-Provider und Erzeugen der PersonViewModel Instanzen 
    public void LoadPersons()
    {
      _persons.Clear();
      List<Person> persons = _dataProvider.GetPersons();
      foreach (Person p in persons)
      {
        PersonViewModel personVM = new PersonViewModel(p);
        personVM.PropertyChanged += PropagateHasChangesAndHasError;
        _persons.Add(personVM);
      }
      UpdateSorting();
      PersonsView.MoveCurrentTo(null);
    }
    
    // Muss vom Window aufgerufen werden, wenn dieses geschlossen wird
    // Gibt false zurck, falls das Fenster nicht geschlossen werden darf
    public bool Close()
    {
      bool ret;

      if (!this.HasChanges && !this.HasError)
      {
        ret = true;
      }
      else if (!this.HasError && this.HasChanges)
      {
        UserFeedbackEventArgs ufe = new UserFeedbackEventArgs("Close", "Save Changes?", MessageBoxButton.YesNoCancel, MessageBoxImage.Question);
        OnUserFeedbackRequired(ufe);
        switch (ufe.Result)
        {
          case MessageBoxResult.Yes:
            HandleSavePersonsExecuted(null, null);
            ret = true;
            break;
          case MessageBoxResult.No:
            ret = true;
            break;
          default:
            ret = false;
            break;
        }
      }
      else
      {
        UserFeedbackEventArgs ufe = new UserFeedbackEventArgs("Close", "Because of errors the changes cannot be saved. Continue without saving?", MessageBoxButton.OKCancel, MessageBoxImage.Exclamation);
        OnUserFeedbackRequired(ufe);
        ret = ufe.Result == MessageBoxResult.OK;
      }
      return ret;
    }

    #endregion

    #region User-Feedback ber MessageBox

    // Das Window muss dieses Event behandeln und die MessageBox anzeigen
    public event EventHandler<UserFeedbackEventArgs> UserFeedbackRequired;

    protected virtual void OnUserFeedbackRequired(UserFeedbackEventArgs e)
    {
      if (UserFeedbackRequired == null) throw new InvalidOperationException("UserFeedbackRequired event must have a listener.");
      UserFeedbackRequired(this, e);
    }

    private void RaiseUserFeedbackRequired(UserFeedbackEventArgs e)
    {
      OnUserFeedbackRequired(e);
    }

    #endregion

    #region Implementation der abstrakten Methoden von ViewModelBase (AcceptChanges, UndoChanges)

    public override void UndoChanges()
    {
      foreach (PersonViewModel personVM in _deletedPersons)
      {
        _persons.Add(personVM);
        personVM.PropertyChanged += PropagateHasChangesAndHasError;
      }
      List<PersonViewModel> addedPersons = new List<PersonViewModel>();
      foreach (PersonViewModel personVM in _persons)
      {
        if (personVM.IsNew)
        {
          addedPersons.Add(personVM);
        }
        else
        {
          personVM.UndoChanges();
        }
      }
      foreach (PersonViewModel personVM in addedPersons)
      {
        personVM.PropertyChanged -= PropagateHasChangesAndHasError;
        _persons.Remove(personVM);
      }
      _deletedPersons.Clear();
    }

    public override void AcceptChanges()
    {
      foreach (PersonViewModel personVM in _persons)
      {
        personVM.AcceptChanges();
      }
      _deletedPersons.Clear();
      HasChanges = false;
    }

    #endregion

    #region Private Hilfsmethoden

    private void SendChangesToModel()
    {
      List<PersonViewModel> newWithoutChanges = new List<PersonViewModel>();
      foreach (PersonViewModel personVM in _persons)
      {
        if (personVM.IsNew && !personVM.HasChanges)
        {
          newWithoutChanges.Add(personVM);
        }
        else
        {
          personVM.SendChangesToModel();
          if (personVM.HasChanges)
          {
            if (personVM.IsNew)
            {
              _dataProvider.InsertPerson(personVM.Person);
            }
            else
            {
              _dataProvider.UpdatePerson(personVM.Person);
            }
          }
        }
      }
      foreach (PersonViewModel personVM in newWithoutChanges)
      {
        _persons.Remove(personVM);
      }
      foreach (PersonViewModel personVM in _deletedPersons)
      {
        if (!personVM.IsNew) _dataProvider.DeletePerson(personVM.Person);
      }
    }

    private void UpdateSorting()
    {
      object current = _personsView.CurrentItem;
      PersonsView.SortDescriptions.Clear();
      PersonsView.SortDescriptions.Add(new SortDescription(this.SortByProperty, ListSortDirection.Ascending));
      _personsView.MoveCurrentTo(current);
    }

    private void InternalDeletePerson(PersonViewModel personVM)
    {
      personVM.PropertyChanged -= PropagateHasChangesAndHasError;
      _persons.Remove(personVM);
      if(!personVM.IsNew) _deletedPersons.Add(personVM);
      UpdateHasChanges();
      UpdateHasError();
    }

    private void InternalInsertPerson()
    {
      PersonViewModel personVM = new PersonViewModel(null);
      personVM.PropertyChanged += PropagateHasChangesAndHasError;
      _persons.Add(personVM);
      _personsView.MoveCurrentTo(personVM);
      personVM.Birthdate.EditValue = string.Empty; // Trick, um die Validierung auszulsen
    }

    private void PropagateHasChangesAndHasError(object sender, PropertyChangedEventArgs e)
    {
      if (e.PropertyName != "HasChanges" && e.PropertyName != "HasError") return;
      ViewModelBase vm = (ViewModelBase)sender;

      if (e.PropertyName == "HasChanges")
      {
        if (vm.HasChanges)
        {
          this.HasChanges = true;
        }
        else
        {
          UpdateHasChanges();
        }
      }
      else if (e.PropertyName == "HasError")
      {
        if (vm.HasError)
        {
          this.HasError = true;
        }
        else
        {
          UpdateHasError();
        }
      }
    }

    private void UpdateHasChanges()
    {
      bool hasChanges = _deletedPersons.Count != 0;
      if (!hasChanges)
      {
        foreach (PersonViewModel personVM in _persons)
        {
          if (personVM.HasChanges)
          {
            hasChanges = true;
            break;
          }
        }
      }
      HasChanges = hasChanges;
    }

    private void UpdateHasError()
    {
      bool hasError = false;
      foreach (PersonViewModel personVM in _persons)
      {
        if (personVM.HasError)
        {
          HasError = true;
          break;
        }
      }
      HasError = hasError;
    }
    #endregion
  }
}
